import { useContext, useEffect, useState, useRef, useCallback } from 'react';
import { StoreContext } from './context';
import { differs } from './differs';
import { raf } from './raf';
var isServer = typeof navigator === 'undefined';

var identity = function identity(x) {
  return x;
};

var immediate = function immediate(fn) {
  return fn();
};

var delayed = function delayed(fn) {
  return raf(fn)();
};

var i = 0;

var nextOrder = function nextOrder() {
  return ++i;
};

export function useSelector(selectorFn, options) {
  if (selectorFn === void 0) {
    selectorFn = identity;
  }

  if (options === void 0) {
    options = {};
  }

  var _options = options,
      _options$sync = _options.sync,
      sync = _options$sync === void 0 ? false : _options$sync,
      _options$pure = _options.pure,
      pure = _options$pure === void 0 ? true : _options$pure,
      _options$observe = _options.observe,
      observe = _options$observe === void 0 ? !isServer : _options$observe;

  var _useContext = useContext(StoreContext),
      atom = _useContext.atom;

  assert(atom, 'No atom found in context, did you forget to wrap your app in <Provider atom={atom} />?'); // cache the schedule and selector functions

  var schedule = useCallback(sync ? immediate : delayed, [sync]);
  var selector = useCallback(selectorFn, options.deps || []); // we use a state to trigger a rerender when relevant atom
  // state changes, we don't store the actual mapped atom state
  // here, because that is only 1 of 2 ways that the component
  // gets rerendered, the other way is being rerended by parent

  var _useState = useState({}),
      rerender = _useState[1]; // keep track of rendering order, this is important for:
  // - correctness – parent must rerender first
  // - performance – parent rerendering children should cancel children's scheduled rerenders


  var order = useRef(); // keep last used props here for diffing upon each change

  var mappedProps = useRef(); // for cancelling scheduled updates in case of parent renders

  var cancelUpdate = useRef(null);

  if (!order.current) {
    order.current = nextOrder();
  } // store current mapped state on each render
  // so we can diff when atom triggers callbacks


  mappedProps.current = selector(atom.get()); // cancel any pending scheduled updates after each render
  // since we just go rerendered by the parent component

  invoke(cancelUpdate);
  useEffect(function observe() {
    if (!observe) return; // very important to check for this, since
    // our observe callback might have been removed
    // from the atom's listeners array while atom is
    // looping over the old list of listener references

    var didUnobserve = false;
    var unobserve = atom.observe(onChange, order.current); // avoid race render/commit phase conditions
    // trigger this to check if atom's state change before
    // we managed to subscribe in this effect

    onChange();

    function onChange() {
      if (didUnobserve) return; // take into account store updates happening in rapid sequence
      // cancel each previously scheduled one and reschedule

      invoke(cancelUpdate); // schedule an update

      cancelUpdate.current = schedule(function scheduledOnChange() {
        cancelUpdate.current = null;
        var nextMappedProps = selector(atom.get());

        if (!pure || differs(mappedProps.current, nextMappedProps)) {
          rerender({});
        }
      });
    }

    return function destroy() {
      didUnobserve = true;
      unobserve();
      invoke(cancelUpdate);
    };
  }, [atom, observe, pure, selector, schedule, order, mappedProps, cancelUpdate, rerender]); // always return fresh mapped props, in case
  // this is a parent rerendering children

  return mappedProps.current;
}
export function useActions() {
  var _useContext2 = useContext(StoreContext),
      atom = _useContext2.atom;

  return atom.actions;
}
export function useDispatch() {
  var _useContext3 = useContext(StoreContext),
      atom = _useContext3.atom;

  return atom.dispatch;
}

function assert(cond, error) {
  if (!cond) {
    throw new Error(error);
  }
}

function invoke(ref) {
  if (ref.current) {
    ref.current();
    ref.current = null;
  }
}